@@ -1,5 +1,6 @@
 
 import gc
+import inspect
 import itertools
 import functools
 
@@ -16,9 +17,9 @@ else:
     import json
 
 if is_utf_console:
-    __vengeance_prefix__ = '\tν: '    # nu: chr(957)
+    __vengeance_prefix__ = '\tν: '    # ν: chr(957), nu
 else:
-    __vengeance_prefix__ = '\tv: '    # ascii
+    __vengeance_prefix__ = '\tv: '    # v: chr(118)
 
 
 def print_runtime(f):
@@ -44,13 +45,14 @@ def print_performance(f=None, repeat=5):
     def performance_wrapper(_f_):
         @functools.wraps(_f_)
         def functools_wrapper(*args, **kwargs):
+
             retv  = None
             best  = None
             worst = None
             total = 0.0
-
             functools_repeat = functools.partial(itertools.repeat, None)
-            was_gc_enabled   = gc.isenabled()
+
+            gc_enabled = gc.isenabled()
             gc.disable()
 
             for _ in functools_repeat(repeat):
@@ -59,31 +61,30 @@ def print_performance(f=None, repeat=5):
                 toc  = default_timer()
                 elapsed = -(tic - toc)
 
+                if best is None: best = elapsed
+                else:            best = min(best, elapsed)
+
+                if worst is None: worst = elapsed
+                else:             worst = max(worst, elapsed)
+
                 total += elapsed
 
-                if best is None:
-                    best = elapsed
-                else:
-                    best = min(best, elapsed)
+            average = total / repeat
+            r = format_header(format_integer(repeat))
 
-                if worst is None:
-                    worst = elapsed
-                else:
-                    worst = max(worst, elapsed)
-
-            s = ('@{}, {}'
+            s = ('@{} {}  trials'
                  '\n        {}best:    {}'
                  '\n        {}average: {}'
                  '\n        {}worst:   {}'
-                 .format(function_name(_f_),
-                         format_header_lite(str(repeat) + ' trials'),
-                         styled('★★★ ', 'blue'),    format_seconds(best),
-                         styled('★★   ', 'blue'), format_seconds(total / repeat),
-                         styled('★    ', 'blue'),    format_seconds(worst)))
+                 .format(function_name(_f_), r,
+                         styled('★★★ ',     'blue'), format_seconds(best),
+                         styled('★★   ', 'blue'), format_seconds(average),
+                         styled('★    ',     'blue'), format_seconds(worst)))
+
             s = vengeance_message(s)
             print(s)
 
-            if was_gc_enabled:
+            if gc_enabled:
                 gc.enable()
 
             return retv
@@ -130,8 +131,11 @@ def vengeance_warning(message,
                       stackframe=None,
                       color=None,
                       effect='bold'):
+    """
+    follow icecream's implementation?
+        call_frame = inspect.currentframe().f_back
+    """
 
-    import inspect
     import traceback
     import warnings
 
@@ -215,7 +219,7 @@ def styled(message,
                     '':          '',
                     None:        ''}
     color_codes = {'grey':           '\x1b[29m',
-                   'white':          '\x1b[30m',
+                   'white':          '\x1b[38;2;255;255;255m',
                    'red':            '\x1b[31m',
                    'green':          '\x1b[32m',
                    'yellow':         '\x1b[33m',
@@ -230,7 +234,7 @@ def styled(message,
                    'bright magenta': '\x1b[95m',
                    'bright cyan':    '\x1b[96m',
                    'black':          '\x1b[97m',
-                   'dark magenta':   '\x1b[38;2;170;50;130m',
+                   'dark magenta':   '\x1b[38;2;180;80;145m',
                    '':               '',
                    None:             ''}
     # endregion
@@ -323,6 +327,31 @@ def object_name(o):
         return o.__class__.__name__
 
 
+def function_parameters(f):
+
+    # region {closure}
+    class param_cls:
+        __slots__ = ('name',
+                     'kind',
+                     'default',
+                     'value')
+
+        def __init__(self, p):
+            self.name    = p.name
+            self.kind    = str(p.kind)
+            self.default = p.default
+            self.value   = p.default
+
+        def __repr__(self):
+            return '{} {}'.format(self.name, self.kind)
+    # endregion
+
+    i_params = inspect.signature(f).parameters
+    n_params = [param_cls(p) for p in i_params.values()]
+
+    return n_params
+
+
 def json_dumps_extended(o, **kwargs):
     kwargs['ensure_ascii'] = kwargs.get('ensure_ascii', False)
     kwargs['indent']       = kwargs.get('indent', 4)
